# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import service as _service
from google.protobuf import service_reflection
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2
from . import databricks_pb2 as databricks__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rservice.proto\x12\x06mlflow\x1a\x15scalapb/scalapb.proto\x1a\x10\x64\x61tabricks.proto\"H\n\x06Metric\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\x0f\n\x04step\x18\x04 \x01(\x03:\x01\x30\"#\n\x05Param\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"f\n\x03Run\x12\x1d\n\x04info\x18\x01 \x01(\x0b\x32\x0f.mlflow.RunInfo\x12\x1d\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x0f.mlflow.RunData\x12!\n\x06inputs\x18\x03 \x01(\x0b\x32\x11.mlflow.RunInputs\"g\n\x07RunData\x12\x1f\n\x07metrics\x18\x01 \x03(\x0b\x32\x0e.mlflow.Metric\x12\x1d\n\x06params\x18\x02 \x03(\x0b\x32\r.mlflow.Param\x12\x1c\n\x04tags\x18\x03 \x03(\x0b\x32\x0e.mlflow.RunTag\"9\n\tRunInputs\x12,\n\x0e\x64\x61taset_inputs\x18\x01 \x03(\x0b\x32\x14.mlflow.DatasetInput\"$\n\x06RunTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"+\n\rExperimentTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"\xdd\x01\n\x07RunInfo\x12\x0e\n\x06run_id\x18\x0f \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x03 \x01(\t\x12\x15\n\rexperiment_id\x18\x02 \x01(\t\x12\x0f\n\x07user_id\x18\x06 \x01(\t\x12!\n\x06status\x18\x07 \x01(\x0e\x32\x11.mlflow.RunStatus\x12\x12\n\nstart_time\x18\x08 \x01(\x03\x12\x10\n\x08\x65nd_time\x18\t \x01(\x03\x12\x14\n\x0c\x61rtifact_uri\x18\r \x01(\t\x12\x17\n\x0flifecycle_stage\x18\x0e \x01(\t\"\xbb\x01\n\nExperiment\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x19\n\x11\x61rtifact_location\x18\x03 \x01(\t\x12\x17\n\x0flifecycle_stage\x18\x04 \x01(\t\x12\x18\n\x10last_update_time\x18\x05 \x01(\x03\x12\x15\n\rcreation_time\x18\x06 \x01(\x03\x12#\n\x04tags\x18\x07 \x03(\x0b\x32\x15.mlflow.ExperimentTag\"V\n\x0c\x44\x61tasetInput\x12\x1e\n\x04tags\x18\x01 \x03(\x0b\x32\x10.mlflow.InputTag\x12&\n\x07\x64\x61taset\x18\x02 \x01(\x0b\x32\x0f.mlflow.DatasetB\x04\xf8\x86\x19\x01\"2\n\x08InputTag\x12\x11\n\x03key\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\"\x85\x01\n\x07\x44\x61taset\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x14\n\x06\x64igest\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x19\n\x0bsource_type\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x12\x14\n\x06source\x18\x04 \x01(\tB\x04\xf8\x86\x19\x01\x12\x0e\n\x06schema\x18\x05 \x01(\t\x12\x0f\n\x07profile\x18\x06 \x01(\t\"\xb6\x01\n\x10\x43reateExperiment\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x19\n\x11\x61rtifact_location\x18\x02 \x01(\t\x12#\n\x04tags\x18\x03 \x03(\x0b\x32\x15.mlflow.ExperimentTag\x1a!\n\x08Response\x12\x15\n\rexperiment_id\x18\x01 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xfe\x01\n\x11SearchExperiments\x12\x13\n\x0bmax_results\x18\x01 \x01(\x03\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x0e\n\x06\x66ilter\x18\x03 \x01(\t\x12\x10\n\x08order_by\x18\x04 \x03(\t\x12#\n\tview_type\x18\x05 \x01(\x0e\x32\x10.mlflow.ViewType\x1aL\n\x08Response\x12\'\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32\x12.mlflow.Experiment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x8d\x01\n\rGetExperiment\x12\x1b\n\rexperiment_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\x32\n\x08Response\x12&\n\nexperiment\x18\x01 \x01(\x0b\x32\x12.mlflow.Experiment:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"h\n\x10\x44\x65leteExperiment\x12\x1b\n\rexperiment_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"i\n\x11RestoreExperiment\x12\x1b\n\rexperiment_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"z\n\x10UpdateExperiment\x12\x1b\n\rexperiment_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x10\n\x08new_name\x18\x02 \x01(\t\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xca\x01\n\tCreateRun\x12\x15\n\rexperiment_id\x18\x01 \x01(\t\x12\x0f\n\x07user_id\x18\x02 \x01(\t\x12\x10\n\x08run_name\x18\x03 \x01(\t\x12\x12\n\nstart_time\x18\x07 \x01(\x03\x12\x1c\n\x04tags\x18\t \x03(\x0b\x32\x0e.mlflow.RunTag\x1a$\n\x08Response\x12\x18\n\x03run\x18\x01 \x01(\x0b\x32\x0b.mlflow.Run:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xd0\x01\n\tUpdateRun\x12\x0e\n\x06run_id\x18\x04 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12!\n\x06status\x18\x02 \x01(\x0e\x32\x11.mlflow.RunStatus\x12\x10\n\x08\x65nd_time\x18\x03 \x01(\x03\x12\x10\n\x08run_name\x18\x05 \x01(\t\x1a-\n\x08Response\x12!\n\x08run_info\x18\x01 \x01(\x0b\x32\x0f.mlflow.RunInfo:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"Z\n\tDeleteRun\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"[\n\nRestoreRun\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb8\x01\n\tLogMetric\x12\x0e\n\x06run_id\x18\x06 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x03 \x01(\x01\x42\x04\xf8\x86\x19\x01\x12\x17\n\ttimestamp\x18\x04 \x01(\x03\x42\x04\xf8\x86\x19\x01\x12\x0f\n\x04step\x18\x05 \x01(\x03:\x01\x30\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x8d\x01\n\x08LogParam\x12\x0e\n\x06run_id\x18\x04 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x90\x01\n\x10SetExperimentTag\x12\x1b\n\rexperiment_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x8b\x01\n\x06SetTag\x12\x0e\n\x06run_id\x18\x04 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x13\n\x05value\x18\x03 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"m\n\tDeleteTag\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x11\n\x03key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"}\n\x06GetRun\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x1a$\n\x08Response\x12\x18\n\x03run\x18\x01 \x01(\x0b\x32\x0b.mlflow.Run:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\x98\x02\n\nSearchRuns\x12\x16\n\x0e\x65xperiment_ids\x18\x01 \x03(\t\x12\x0e\n\x06\x66ilter\x18\x04 \x01(\t\x12\x34\n\rrun_view_type\x18\x03 \x01(\x0e\x32\x10.mlflow.ViewType:\x0b\x41\x43TIVE_ONLY\x12\x19\n\x0bmax_results\x18\x05 \x01(\x05:\x04\x31\x30\x30\x30\x12\x10\n\x08order_by\x18\x06 \x03(\t\x12\x12\n\npage_token\x18\x07 \x01(\t\x1a>\n\x08Response\x12\x19\n\x04runs\x18\x01 \x03(\x0b\x32\x0b.mlflow.Run\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xd8\x01\n\rListArtifacts\x12\x0e\n\x06run_id\x18\x03 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x12\n\npage_token\x18\x04 \x01(\t\x1aV\n\x08Response\x12\x10\n\x08root_uri\x18\x01 \x01(\t\x12\x1f\n\x05\x66iles\x18\x02 \x03(\x0b\x32\x10.mlflow.FileInfo\x12\x17\n\x0fnext_page_token\x18\x03 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\";\n\x08\x46ileInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x0e\n\x06is_dir\x18\x02 \x01(\x08\x12\x11\n\tfile_size\x18\x03 \x01(\x03\"\xea\x01\n\x10GetMetricHistory\x12\x0e\n\x06run_id\x18\x03 \x01(\t\x12\x10\n\x08run_uuid\x18\x01 \x01(\t\x12\x18\n\nmetric_key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x12\n\npage_token\x18\x04 \x01(\t\x12\x13\n\x0bmax_results\x18\x05 \x01(\x05\x1a\x44\n\x08Response\x12\x1f\n\x07metrics\x18\x01 \x03(\x0b\x32\x0e.mlflow.Metric\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"a\n\x0fMetricWithRunId\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12\x0f\n\x04step\x18\x04 \x01(\x03:\x01\x30\x12\x0e\n\x06run_id\x18\x05 \x01(\t\"\x9b\x02\n\x1cGetMetricHistoryBulkInterval\x12\x0f\n\x07run_ids\x18\x01 \x03(\t\x12\x18\n\nmetric_key\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x12\x12\n\nstart_step\x18\x03 \x01(\x05\x12\x10\n\x08\x65nd_step\x18\x04 \x01(\x05\x12\x13\n\x0bmax_results\x18\x05 \x01(\x05\x1a\x34\n\x08Response\x12(\n\x07metrics\x18\x01 \x03(\x0b\x32\x17.mlflow.MetricWithRunId:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage\"\xb1\x01\n\x08LogBatch\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x1f\n\x07metrics\x18\x02 \x03(\x0b\x32\x0e.mlflow.Metric\x12\x1d\n\x06params\x18\x03 \x03(\x0b\x32\r.mlflow.Param\x12\x1c\n\x04tags\x18\x04 \x03(\x0b\x32\x0e.mlflow.RunTag\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"g\n\x08LogModel\x12\x0e\n\x06run_id\x18\x01 \x01(\t\x12\x12\n\nmodel_json\x18\x02 \x01(\t\x1a\n\n\x08Response:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\"\xb6\x01\n\tLogInputs\x12\x14\n\x06run_id\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12&\n\x08\x64\x61tasets\x18\x02 \x03(\x0b\x32\x14.mlflow.DatasetInput\x1a\n\n\x08Response:_\xe2?(\n&com.databricks.rpc.RPC[$this.Response]\xe2?1\n/com.databricks.mlflow.api.MlflowTrackingMessage\"\x95\x01\n\x13GetExperimentByName\x12\x1d\n\x0f\x65xperiment_name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x1a\x32\n\x08Response\x12&\n\nexperiment\x18\x01 \x01(\x0b\x32\x12.mlflow.Experiment:+\xe2?(\n&com.databricks.rpc.RPC[$this.Response]*6\n\x08ViewType\x12\x0f\n\x0b\x41\x43TIVE_ONLY\x10\x01\x12\x10\n\x0c\x44\x45LETED_ONLY\x10\x02\x12\x07\n\x03\x41LL\x10\x03*I\n\nSourceType\x12\x0c\n\x08NOTEBOOK\x10\x01\x12\x07\n\x03JOB\x10\x02\x12\x0b\n\x07PROJECT\x10\x03\x12\t\n\x05LOCAL\x10\x04\x12\x0c\n\x07UNKNOWN\x10\xe8\x07*M\n\tRunStatus\x12\x0b\n\x07RUNNING\x10\x01\x12\r\n\tSCHEDULED\x10\x02\x12\x0c\n\x08\x46INISHED\x10\x03\x12\n\n\x06\x46\x41ILED\x10\x04\x12\n\n\x06KILLED\x10\x05\x32\xb6\x19\n\rMlflowService\x12\xa6\x01\n\x13getExperimentByName\x12\x1b.mlflow.GetExperimentByName\x1a$.mlflow.GetExperimentByName.Response\"L\xf2\x86\x19H\n,\n\x03GET\x12\x1f/mlflow/experiments/get-by-name\x1a\x04\x08\x02\x10\x00\x10\x01*\x16Get Experiment By Name\x12\x94\x01\n\x10\x63reateExperiment\x12\x18.mlflow.CreateExperiment\x1a!.mlflow.CreateExperiment.Response\"C\xf2\x86\x19?\n(\n\x04POST\x12\x1a/mlflow/experiments/create\x1a\x04\x08\x02\x10\x00\x10\x01*\x11\x43reate Experiment\x12\xc1\x01\n\x11searchExperiments\x12\x19.mlflow.SearchExperiments\x1a\".mlflow.SearchExperiments.Response\"m\xf2\x86\x19i\n(\n\x04POST\x12\x1a/mlflow/experiments/search\x1a\x04\x08\x02\x10\x00\n\'\n\x03GET\x12\x1a/mlflow/experiments/search\x1a\x04\x08\x02\x10\x00\x10\x01*\x12Search Experiments\x12\x88\x01\n\rgetExperiment\x12\x15.mlflow.GetExperiment\x1a\x1e.mlflow.GetExperiment.Response\"@\xf2\x86\x19\x38\n$\n\x03GET\x12\x17/mlflow/experiments/get\x1a\x04\x08\x02\x10\x00\x10\x01*\x0eGet Experiment\xba\x8c\x19\x00\x12\x94\x01\n\x10\x64\x65leteExperiment\x12\x18.mlflow.DeleteExperiment\x1a!.mlflow.DeleteExperiment.Response\"C\xf2\x86\x19?\n(\n\x04POST\x12\x1a/mlflow/experiments/delete\x1a\x04\x08\x02\x10\x00\x10\x01*\x11\x44\x65lete Experiment\x12\x99\x01\n\x11restoreExperiment\x12\x19.mlflow.RestoreExperiment\x1a\".mlflow.RestoreExperiment.Response\"E\xf2\x86\x19\x41\n)\n\x04POST\x12\x1b/mlflow/experiments/restore\x1a\x04\x08\x02\x10\x00\x10\x01*\x12Restore Experiment\x12\x94\x01\n\x10updateExperiment\x12\x18.mlflow.UpdateExperiment\x1a!.mlflow.UpdateExperiment.Response\"C\xf2\x86\x19?\n(\n\x04POST\x12\x1a/mlflow/experiments/update\x1a\x04\x08\x02\x10\x00\x10\x01*\x11Update Experiment\x12q\n\tcreateRun\x12\x11.mlflow.CreateRun\x1a\x1a.mlflow.CreateRun.Response\"5\xf2\x86\x19\x31\n!\n\x04POST\x12\x13/mlflow/runs/create\x1a\x04\x08\x02\x10\x00\x10\x01*\nCreate Run\x12q\n\tupdateRun\x12\x11.mlflow.UpdateRun\x1a\x1a.mlflow.UpdateRun.Response\"5\xf2\x86\x19\x31\n!\n\x04POST\x12\x13/mlflow/runs/update\x1a\x04\x08\x02\x10\x00\x10\x01*\nUpdate Run\x12q\n\tdeleteRun\x12\x11.mlflow.DeleteRun\x1a\x1a.mlflow.DeleteRun.Response\"5\xf2\x86\x19\x31\n!\n\x04POST\x12\x13/mlflow/runs/delete\x1a\x04\x08\x02\x10\x00\x10\x01*\nDelete Run\x12v\n\nrestoreRun\x12\x12.mlflow.RestoreRun\x1a\x1b.mlflow.RestoreRun.Response\"7\xf2\x86\x19\x33\n\"\n\x04POST\x12\x14/mlflow/runs/restore\x1a\x04\x08\x02\x10\x00\x10\x01*\x0bRestore Run\x12u\n\tlogMetric\x12\x11.mlflow.LogMetric\x1a\x1a.mlflow.LogMetric.Response\"9\xf2\x86\x19\x35\n%\n\x04POST\x12\x17/mlflow/runs/log-metric\x1a\x04\x08\x02\x10\x00\x10\x01*\nLog Metric\x12t\n\x08logParam\x12\x10.mlflow.LogParam\x1a\x19.mlflow.LogParam.Response\";\xf2\x86\x19\x37\n(\n\x04POST\x12\x1a/mlflow/runs/log-parameter\x1a\x04\x08\x02\x10\x00\x10\x01*\tLog Param\x12\xa1\x01\n\x10setExperimentTag\x12\x18.mlflow.SetExperimentTag\x1a!.mlflow.SetExperimentTag.Response\"P\xf2\x86\x19L\n4\n\x04POST\x12&/mlflow/experiments/set-experiment-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x12Set Experiment Tag\x12\x66\n\x06setTag\x12\x0e.mlflow.SetTag\x1a\x17.mlflow.SetTag.Response\"3\xf2\x86\x19/\n\"\n\x04POST\x12\x14/mlflow/runs/set-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\x07Set Tag\x12u\n\tdeleteTag\x12\x11.mlflow.DeleteTag\x1a\x1a.mlflow.DeleteTag.Response\"9\xf2\x86\x19\x35\n%\n\x04POST\x12\x17/mlflow/runs/delete-tag\x1a\x04\x08\x02\x10\x00\x10\x01*\nDelete Tag\x12\x65\n\x06getRun\x12\x0e.mlflow.GetRun\x1a\x17.mlflow.GetRun.Response\"2\xf2\x86\x19*\n\x1d\n\x03GET\x12\x10/mlflow/runs/get\x1a\x04\x08\x02\x10\x00\x10\x01*\x07Get Run\xba\x8c\x19\x00\x12u\n\nsearchRuns\x12\x12.mlflow.SearchRuns\x1a\x1b.mlflow.SearchRuns.Response\"6\xf2\x86\x19\x32\n!\n\x04POST\x12\x13/mlflow/runs/search\x1a\x04\x08\x02\x10\x00\x10\x01*\x0bSearch Runs\x12\x83\x01\n\rlistArtifacts\x12\x15.mlflow.ListArtifacts\x1a\x1e.mlflow.ListArtifacts.Response\";\xf2\x86\x19\x37\n#\n\x03GET\x12\x16/mlflow/artifacts/list\x1a\x04\x08\x02\x10\x00\x10\x01*\x0eList Artifacts\x12\x95\x01\n\x10getMetricHistory\x12\x18.mlflow.GetMetricHistory\x1a!.mlflow.GetMetricHistory.Response\"D\xf2\x86\x19@\n(\n\x03GET\x12\x1b/mlflow/metrics/get-history\x1a\x04\x08\x02\x10\x00\x10\x01*\x12Get Metric History\x12\xb3\x01\n\x1cgetMetricHistoryBulkInterval\x12$.mlflow.GetMetricHistoryBulkInterval\x1a-.mlflow.GetMetricHistoryBulkInterval.Response\">\xf2\x86\x19:\n6\n\x03GET\x12)/mlflow/metrics/get-history-bulk-interval\x1a\x04\x08\x02\x10\x0b\x10\x03\x12p\n\x08logBatch\x12\x10.mlflow.LogBatch\x1a\x19.mlflow.LogBatch.Response\"7\xf2\x86\x19\x33\n$\n\x04POST\x12\x16/mlflow/runs/log-batch\x1a\x04\x08\x02\x10\x00\x10\x01*\tLog Batch\x12p\n\x08logModel\x12\x10.mlflow.LogModel\x1a\x19.mlflow.LogModel.Response\"7\xf2\x86\x19\x33\n$\n\x04POST\x12\x16/mlflow/runs/log-model\x1a\x04\x08\x02\x10\x00\x10\x01*\tLog Model\x12u\n\tlogInputs\x12\x11.mlflow.LogInputs\x1a\x1a.mlflow.LogInputs.Response\"9\xf2\x86\x19\x35\n%\n\x04POST\x12\x17/mlflow/runs/log-inputs\x1a\x04\x08\x02\x10\x00\x10\x01*\nLog InputsB\x1e\n\x14org.mlflow.api.proto\x90\x01\x01\xe2?\x02\x10\x01')

_VIEWTYPE = DESCRIPTOR.enum_types_by_name['ViewType']
ViewType = enum_type_wrapper.EnumTypeWrapper(_VIEWTYPE)
_SOURCETYPE = DESCRIPTOR.enum_types_by_name['SourceType']
SourceType = enum_type_wrapper.EnumTypeWrapper(_SOURCETYPE)
_RUNSTATUS = DESCRIPTOR.enum_types_by_name['RunStatus']
RunStatus = enum_type_wrapper.EnumTypeWrapper(_RUNSTATUS)
ACTIVE_ONLY = 1
DELETED_ONLY = 2
ALL = 3
NOTEBOOK = 1
JOB = 2
PROJECT = 3
LOCAL = 4
UNKNOWN = 1000
RUNNING = 1
SCHEDULED = 2
FINISHED = 3
FAILED = 4
KILLED = 5


_METRIC = DESCRIPTOR.message_types_by_name['Metric']
_PARAM = DESCRIPTOR.message_types_by_name['Param']
_RUN = DESCRIPTOR.message_types_by_name['Run']
_RUNDATA = DESCRIPTOR.message_types_by_name['RunData']
_RUNINPUTS = DESCRIPTOR.message_types_by_name['RunInputs']
_RUNTAG = DESCRIPTOR.message_types_by_name['RunTag']
_EXPERIMENTTAG = DESCRIPTOR.message_types_by_name['ExperimentTag']
_RUNINFO = DESCRIPTOR.message_types_by_name['RunInfo']
_EXPERIMENT = DESCRIPTOR.message_types_by_name['Experiment']
_DATASETINPUT = DESCRIPTOR.message_types_by_name['DatasetInput']
_INPUTTAG = DESCRIPTOR.message_types_by_name['InputTag']
_DATASET = DESCRIPTOR.message_types_by_name['Dataset']
_CREATEEXPERIMENT = DESCRIPTOR.message_types_by_name['CreateExperiment']
_CREATEEXPERIMENT_RESPONSE = _CREATEEXPERIMENT.nested_types_by_name['Response']
_SEARCHEXPERIMENTS = DESCRIPTOR.message_types_by_name['SearchExperiments']
_SEARCHEXPERIMENTS_RESPONSE = _SEARCHEXPERIMENTS.nested_types_by_name['Response']
_GETEXPERIMENT = DESCRIPTOR.message_types_by_name['GetExperiment']
_GETEXPERIMENT_RESPONSE = _GETEXPERIMENT.nested_types_by_name['Response']
_DELETEEXPERIMENT = DESCRIPTOR.message_types_by_name['DeleteExperiment']
_DELETEEXPERIMENT_RESPONSE = _DELETEEXPERIMENT.nested_types_by_name['Response']
_RESTOREEXPERIMENT = DESCRIPTOR.message_types_by_name['RestoreExperiment']
_RESTOREEXPERIMENT_RESPONSE = _RESTOREEXPERIMENT.nested_types_by_name['Response']
_UPDATEEXPERIMENT = DESCRIPTOR.message_types_by_name['UpdateExperiment']
_UPDATEEXPERIMENT_RESPONSE = _UPDATEEXPERIMENT.nested_types_by_name['Response']
_CREATERUN = DESCRIPTOR.message_types_by_name['CreateRun']
_CREATERUN_RESPONSE = _CREATERUN.nested_types_by_name['Response']
_UPDATERUN = DESCRIPTOR.message_types_by_name['UpdateRun']
_UPDATERUN_RESPONSE = _UPDATERUN.nested_types_by_name['Response']
_DELETERUN = DESCRIPTOR.message_types_by_name['DeleteRun']
_DELETERUN_RESPONSE = _DELETERUN.nested_types_by_name['Response']
_RESTORERUN = DESCRIPTOR.message_types_by_name['RestoreRun']
_RESTORERUN_RESPONSE = _RESTORERUN.nested_types_by_name['Response']
_LOGMETRIC = DESCRIPTOR.message_types_by_name['LogMetric']
_LOGMETRIC_RESPONSE = _LOGMETRIC.nested_types_by_name['Response']
_LOGPARAM = DESCRIPTOR.message_types_by_name['LogParam']
_LOGPARAM_RESPONSE = _LOGPARAM.nested_types_by_name['Response']
_SETEXPERIMENTTAG = DESCRIPTOR.message_types_by_name['SetExperimentTag']
_SETEXPERIMENTTAG_RESPONSE = _SETEXPERIMENTTAG.nested_types_by_name['Response']
_SETTAG = DESCRIPTOR.message_types_by_name['SetTag']
_SETTAG_RESPONSE = _SETTAG.nested_types_by_name['Response']
_DELETETAG = DESCRIPTOR.message_types_by_name['DeleteTag']
_DELETETAG_RESPONSE = _DELETETAG.nested_types_by_name['Response']
_GETRUN = DESCRIPTOR.message_types_by_name['GetRun']
_GETRUN_RESPONSE = _GETRUN.nested_types_by_name['Response']
_SEARCHRUNS = DESCRIPTOR.message_types_by_name['SearchRuns']
_SEARCHRUNS_RESPONSE = _SEARCHRUNS.nested_types_by_name['Response']
_LISTARTIFACTS = DESCRIPTOR.message_types_by_name['ListArtifacts']
_LISTARTIFACTS_RESPONSE = _LISTARTIFACTS.nested_types_by_name['Response']
_FILEINFO = DESCRIPTOR.message_types_by_name['FileInfo']
_GETMETRICHISTORY = DESCRIPTOR.message_types_by_name['GetMetricHistory']
_GETMETRICHISTORY_RESPONSE = _GETMETRICHISTORY.nested_types_by_name['Response']
_METRICWITHRUNID = DESCRIPTOR.message_types_by_name['MetricWithRunId']
_GETMETRICHISTORYBULKINTERVAL = DESCRIPTOR.message_types_by_name['GetMetricHistoryBulkInterval']
_GETMETRICHISTORYBULKINTERVAL_RESPONSE = _GETMETRICHISTORYBULKINTERVAL.nested_types_by_name['Response']
_LOGBATCH = DESCRIPTOR.message_types_by_name['LogBatch']
_LOGBATCH_RESPONSE = _LOGBATCH.nested_types_by_name['Response']
_LOGMODEL = DESCRIPTOR.message_types_by_name['LogModel']
_LOGMODEL_RESPONSE = _LOGMODEL.nested_types_by_name['Response']
_LOGINPUTS = DESCRIPTOR.message_types_by_name['LogInputs']
_LOGINPUTS_RESPONSE = _LOGINPUTS.nested_types_by_name['Response']
_GETEXPERIMENTBYNAME = DESCRIPTOR.message_types_by_name['GetExperimentByName']
_GETEXPERIMENTBYNAME_RESPONSE = _GETEXPERIMENTBYNAME.nested_types_by_name['Response']
Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.Metric)
  })
_sym_db.RegisterMessage(Metric)

Param = _reflection.GeneratedProtocolMessageType('Param', (_message.Message,), {
  'DESCRIPTOR' : _PARAM,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.Param)
  })
_sym_db.RegisterMessage(Param)

Run = _reflection.GeneratedProtocolMessageType('Run', (_message.Message,), {
  'DESCRIPTOR' : _RUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.Run)
  })
_sym_db.RegisterMessage(Run)

RunData = _reflection.GeneratedProtocolMessageType('RunData', (_message.Message,), {
  'DESCRIPTOR' : _RUNDATA,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RunData)
  })
_sym_db.RegisterMessage(RunData)

RunInputs = _reflection.GeneratedProtocolMessageType('RunInputs', (_message.Message,), {
  'DESCRIPTOR' : _RUNINPUTS,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RunInputs)
  })
_sym_db.RegisterMessage(RunInputs)

RunTag = _reflection.GeneratedProtocolMessageType('RunTag', (_message.Message,), {
  'DESCRIPTOR' : _RUNTAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RunTag)
  })
_sym_db.RegisterMessage(RunTag)

ExperimentTag = _reflection.GeneratedProtocolMessageType('ExperimentTag', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENTTAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.ExperimentTag)
  })
_sym_db.RegisterMessage(ExperimentTag)

RunInfo = _reflection.GeneratedProtocolMessageType('RunInfo', (_message.Message,), {
  'DESCRIPTOR' : _RUNINFO,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RunInfo)
  })
_sym_db.RegisterMessage(RunInfo)

Experiment = _reflection.GeneratedProtocolMessageType('Experiment', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.Experiment)
  })
_sym_db.RegisterMessage(Experiment)

DatasetInput = _reflection.GeneratedProtocolMessageType('DatasetInput', (_message.Message,), {
  'DESCRIPTOR' : _DATASETINPUT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DatasetInput)
  })
_sym_db.RegisterMessage(DatasetInput)

InputTag = _reflection.GeneratedProtocolMessageType('InputTag', (_message.Message,), {
  'DESCRIPTOR' : _INPUTTAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.InputTag)
  })
_sym_db.RegisterMessage(InputTag)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.Dataset)
  })
_sym_db.RegisterMessage(Dataset)

CreateExperiment = _reflection.GeneratedProtocolMessageType('CreateExperiment', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _CREATEEXPERIMENT_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.CreateExperiment.Response)
    })
  ,
  'DESCRIPTOR' : _CREATEEXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.CreateExperiment)
  })
_sym_db.RegisterMessage(CreateExperiment)
_sym_db.RegisterMessage(CreateExperiment.Response)

SearchExperiments = _reflection.GeneratedProtocolMessageType('SearchExperiments', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHEXPERIMENTS_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SearchExperiments.Response)
    })
  ,
  'DESCRIPTOR' : _SEARCHEXPERIMENTS,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SearchExperiments)
  })
_sym_db.RegisterMessage(SearchExperiments)
_sym_db.RegisterMessage(SearchExperiments.Response)

GetExperiment = _reflection.GeneratedProtocolMessageType('GetExperiment', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETEXPERIMENT_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetExperiment.Response)
    })
  ,
  'DESCRIPTOR' : _GETEXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetExperiment)
  })
_sym_db.RegisterMessage(GetExperiment)
_sym_db.RegisterMessage(GetExperiment.Response)

DeleteExperiment = _reflection.GeneratedProtocolMessageType('DeleteExperiment', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETEEXPERIMENT_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteExperiment.Response)
    })
  ,
  'DESCRIPTOR' : _DELETEEXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteExperiment)
  })
_sym_db.RegisterMessage(DeleteExperiment)
_sym_db.RegisterMessage(DeleteExperiment.Response)

RestoreExperiment = _reflection.GeneratedProtocolMessageType('RestoreExperiment', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _RESTOREEXPERIMENT_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.RestoreExperiment.Response)
    })
  ,
  'DESCRIPTOR' : _RESTOREEXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RestoreExperiment)
  })
_sym_db.RegisterMessage(RestoreExperiment)
_sym_db.RegisterMessage(RestoreExperiment.Response)

UpdateExperiment = _reflection.GeneratedProtocolMessageType('UpdateExperiment', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEEXPERIMENT_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.UpdateExperiment.Response)
    })
  ,
  'DESCRIPTOR' : _UPDATEEXPERIMENT,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.UpdateExperiment)
  })
_sym_db.RegisterMessage(UpdateExperiment)
_sym_db.RegisterMessage(UpdateExperiment.Response)

CreateRun = _reflection.GeneratedProtocolMessageType('CreateRun', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _CREATERUN_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.CreateRun.Response)
    })
  ,
  'DESCRIPTOR' : _CREATERUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.CreateRun)
  })
_sym_db.RegisterMessage(CreateRun)
_sym_db.RegisterMessage(CreateRun.Response)

UpdateRun = _reflection.GeneratedProtocolMessageType('UpdateRun', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _UPDATERUN_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.UpdateRun.Response)
    })
  ,
  'DESCRIPTOR' : _UPDATERUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.UpdateRun)
  })
_sym_db.RegisterMessage(UpdateRun)
_sym_db.RegisterMessage(UpdateRun.Response)

DeleteRun = _reflection.GeneratedProtocolMessageType('DeleteRun', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETERUN_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteRun.Response)
    })
  ,
  'DESCRIPTOR' : _DELETERUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteRun)
  })
_sym_db.RegisterMessage(DeleteRun)
_sym_db.RegisterMessage(DeleteRun.Response)

RestoreRun = _reflection.GeneratedProtocolMessageType('RestoreRun', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _RESTORERUN_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.RestoreRun.Response)
    })
  ,
  'DESCRIPTOR' : _RESTORERUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.RestoreRun)
  })
_sym_db.RegisterMessage(RestoreRun)
_sym_db.RegisterMessage(RestoreRun.Response)

LogMetric = _reflection.GeneratedProtocolMessageType('LogMetric', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LOGMETRIC_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.LogMetric.Response)
    })
  ,
  'DESCRIPTOR' : _LOGMETRIC,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.LogMetric)
  })
_sym_db.RegisterMessage(LogMetric)
_sym_db.RegisterMessage(LogMetric.Response)

LogParam = _reflection.GeneratedProtocolMessageType('LogParam', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LOGPARAM_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.LogParam.Response)
    })
  ,
  'DESCRIPTOR' : _LOGPARAM,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.LogParam)
  })
_sym_db.RegisterMessage(LogParam)
_sym_db.RegisterMessage(LogParam.Response)

SetExperimentTag = _reflection.GeneratedProtocolMessageType('SetExperimentTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SETEXPERIMENTTAG_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SetExperimentTag.Response)
    })
  ,
  'DESCRIPTOR' : _SETEXPERIMENTTAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SetExperimentTag)
  })
_sym_db.RegisterMessage(SetExperimentTag)
_sym_db.RegisterMessage(SetExperimentTag.Response)

SetTag = _reflection.GeneratedProtocolMessageType('SetTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SETTAG_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SetTag.Response)
    })
  ,
  'DESCRIPTOR' : _SETTAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SetTag)
  })
_sym_db.RegisterMessage(SetTag)
_sym_db.RegisterMessage(SetTag.Response)

DeleteTag = _reflection.GeneratedProtocolMessageType('DeleteTag', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _DELETETAG_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DeleteTag.Response)
    })
  ,
  'DESCRIPTOR' : _DELETETAG,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.DeleteTag)
  })
_sym_db.RegisterMessage(DeleteTag)
_sym_db.RegisterMessage(DeleteTag.Response)

GetRun = _reflection.GeneratedProtocolMessageType('GetRun', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETRUN_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetRun.Response)
    })
  ,
  'DESCRIPTOR' : _GETRUN,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetRun)
  })
_sym_db.RegisterMessage(GetRun)
_sym_db.RegisterMessage(GetRun.Response)

SearchRuns = _reflection.GeneratedProtocolMessageType('SearchRuns', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHRUNS_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.SearchRuns.Response)
    })
  ,
  'DESCRIPTOR' : _SEARCHRUNS,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.SearchRuns)
  })
_sym_db.RegisterMessage(SearchRuns)
_sym_db.RegisterMessage(SearchRuns.Response)

ListArtifacts = _reflection.GeneratedProtocolMessageType('ListArtifacts', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LISTARTIFACTS_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.ListArtifacts.Response)
    })
  ,
  'DESCRIPTOR' : _LISTARTIFACTS,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.ListArtifacts)
  })
_sym_db.RegisterMessage(ListArtifacts)
_sym_db.RegisterMessage(ListArtifacts.Response)

FileInfo = _reflection.GeneratedProtocolMessageType('FileInfo', (_message.Message,), {
  'DESCRIPTOR' : _FILEINFO,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.FileInfo)
  })
_sym_db.RegisterMessage(FileInfo)

GetMetricHistory = _reflection.GeneratedProtocolMessageType('GetMetricHistory', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETMETRICHISTORY_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetMetricHistory.Response)
    })
  ,
  'DESCRIPTOR' : _GETMETRICHISTORY,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetMetricHistory)
  })
_sym_db.RegisterMessage(GetMetricHistory)
_sym_db.RegisterMessage(GetMetricHistory.Response)

MetricWithRunId = _reflection.GeneratedProtocolMessageType('MetricWithRunId', (_message.Message,), {
  'DESCRIPTOR' : _METRICWITHRUNID,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.MetricWithRunId)
  })
_sym_db.RegisterMessage(MetricWithRunId)

GetMetricHistoryBulkInterval = _reflection.GeneratedProtocolMessageType('GetMetricHistoryBulkInterval', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETMETRICHISTORYBULKINTERVAL_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetMetricHistoryBulkInterval.Response)
    })
  ,
  'DESCRIPTOR' : _GETMETRICHISTORYBULKINTERVAL,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetMetricHistoryBulkInterval)
  })
_sym_db.RegisterMessage(GetMetricHistoryBulkInterval)
_sym_db.RegisterMessage(GetMetricHistoryBulkInterval.Response)

LogBatch = _reflection.GeneratedProtocolMessageType('LogBatch', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LOGBATCH_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.LogBatch.Response)
    })
  ,
  'DESCRIPTOR' : _LOGBATCH,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.LogBatch)
  })
_sym_db.RegisterMessage(LogBatch)
_sym_db.RegisterMessage(LogBatch.Response)

LogModel = _reflection.GeneratedProtocolMessageType('LogModel', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LOGMODEL_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.LogModel.Response)
    })
  ,
  'DESCRIPTOR' : _LOGMODEL,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.LogModel)
  })
_sym_db.RegisterMessage(LogModel)
_sym_db.RegisterMessage(LogModel.Response)

LogInputs = _reflection.GeneratedProtocolMessageType('LogInputs', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _LOGINPUTS_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.LogInputs.Response)
    })
  ,
  'DESCRIPTOR' : _LOGINPUTS,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.LogInputs)
  })
_sym_db.RegisterMessage(LogInputs)
_sym_db.RegisterMessage(LogInputs.Response)

GetExperimentByName = _reflection.GeneratedProtocolMessageType('GetExperimentByName', (_message.Message,), {

  'Response' : _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), {
    'DESCRIPTOR' : _GETEXPERIMENTBYNAME_RESPONSE,
    '__module__' : 'service_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.GetExperimentByName.Response)
    })
  ,
  'DESCRIPTOR' : _GETEXPERIMENTBYNAME,
  '__module__' : 'service_pb2'
  # @@protoc_insertion_point(class_scope:mlflow.GetExperimentByName)
  })
_sym_db.RegisterMessage(GetExperimentByName)
_sym_db.RegisterMessage(GetExperimentByName.Response)

_MLFLOWSERVICE = DESCRIPTOR.services_by_name['MlflowService']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024org.mlflow.api.proto\220\001\001\342?\002\020\001'
  _DATASETINPUT.fields_by_name['dataset']._options = None
  _DATASETINPUT.fields_by_name['dataset']._serialized_options = b'\370\206\031\001'
  _INPUTTAG.fields_by_name['key']._options = None
  _INPUTTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _INPUTTAG.fields_by_name['value']._options = None
  _INPUTTAG.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _DATASET.fields_by_name['name']._options = None
  _DATASET.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _DATASET.fields_by_name['digest']._options = None
  _DATASET.fields_by_name['digest']._serialized_options = b'\370\206\031\001'
  _DATASET.fields_by_name['source_type']._options = None
  _DATASET.fields_by_name['source_type']._serialized_options = b'\370\206\031\001'
  _DATASET.fields_by_name['source']._options = None
  _DATASET.fields_by_name['source']._serialized_options = b'\370\206\031\001'
  _CREATEEXPERIMENT.fields_by_name['name']._options = None
  _CREATEEXPERIMENT.fields_by_name['name']._serialized_options = b'\370\206\031\001'
  _CREATEEXPERIMENT._options = None
  _CREATEEXPERIMENT._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SEARCHEXPERIMENTS._options = None
  _SEARCHEXPERIMENTS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETEXPERIMENT.fields_by_name['experiment_id']._options = None
  _GETEXPERIMENT.fields_by_name['experiment_id']._serialized_options = b'\370\206\031\001'
  _GETEXPERIMENT._options = None
  _GETEXPERIMENT._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETEEXPERIMENT.fields_by_name['experiment_id']._options = None
  _DELETEEXPERIMENT.fields_by_name['experiment_id']._serialized_options = b'\370\206\031\001'
  _DELETEEXPERIMENT._options = None
  _DELETEEXPERIMENT._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _RESTOREEXPERIMENT.fields_by_name['experiment_id']._options = None
  _RESTOREEXPERIMENT.fields_by_name['experiment_id']._serialized_options = b'\370\206\031\001'
  _RESTOREEXPERIMENT._options = None
  _RESTOREEXPERIMENT._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _UPDATEEXPERIMENT.fields_by_name['experiment_id']._options = None
  _UPDATEEXPERIMENT.fields_by_name['experiment_id']._serialized_options = b'\370\206\031\001'
  _UPDATEEXPERIMENT._options = None
  _UPDATEEXPERIMENT._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _CREATERUN._options = None
  _CREATERUN._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _UPDATERUN._options = None
  _UPDATERUN._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETERUN.fields_by_name['run_id']._options = None
  _DELETERUN.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _DELETERUN._options = None
  _DELETERUN._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _RESTORERUN.fields_by_name['run_id']._options = None
  _RESTORERUN.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _RESTORERUN._options = None
  _RESTORERUN._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _LOGMETRIC.fields_by_name['key']._options = None
  _LOGMETRIC.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _LOGMETRIC.fields_by_name['value']._options = None
  _LOGMETRIC.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _LOGMETRIC.fields_by_name['timestamp']._options = None
  _LOGMETRIC.fields_by_name['timestamp']._serialized_options = b'\370\206\031\001'
  _LOGMETRIC._options = None
  _LOGMETRIC._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _LOGPARAM.fields_by_name['key']._options = None
  _LOGPARAM.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _LOGPARAM.fields_by_name['value']._options = None
  _LOGPARAM.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _LOGPARAM._options = None
  _LOGPARAM._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SETEXPERIMENTTAG.fields_by_name['experiment_id']._options = None
  _SETEXPERIMENTTAG.fields_by_name['experiment_id']._serialized_options = b'\370\206\031\001'
  _SETEXPERIMENTTAG.fields_by_name['key']._options = None
  _SETEXPERIMENTTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _SETEXPERIMENTTAG.fields_by_name['value']._options = None
  _SETEXPERIMENTTAG.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _SETEXPERIMENTTAG._options = None
  _SETEXPERIMENTTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SETTAG.fields_by_name['key']._options = None
  _SETTAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _SETTAG.fields_by_name['value']._options = None
  _SETTAG.fields_by_name['value']._serialized_options = b'\370\206\031\001'
  _SETTAG._options = None
  _SETTAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _DELETETAG.fields_by_name['run_id']._options = None
  _DELETETAG.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _DELETETAG.fields_by_name['key']._options = None
  _DELETETAG.fields_by_name['key']._serialized_options = b'\370\206\031\001'
  _DELETETAG._options = None
  _DELETETAG._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETRUN._options = None
  _GETRUN._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _SEARCHRUNS._options = None
  _SEARCHRUNS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _LISTARTIFACTS._options = None
  _LISTARTIFACTS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETMETRICHISTORY.fields_by_name['metric_key']._options = None
  _GETMETRICHISTORY.fields_by_name['metric_key']._serialized_options = b'\370\206\031\001'
  _GETMETRICHISTORY._options = None
  _GETMETRICHISTORY._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _GETMETRICHISTORYBULKINTERVAL.fields_by_name['metric_key']._options = None
  _GETMETRICHISTORYBULKINTERVAL.fields_by_name['metric_key']._serialized_options = b'\370\206\031\001'
  _GETMETRICHISTORYBULKINTERVAL._options = None
  _GETMETRICHISTORYBULKINTERVAL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _LOGBATCH._options = None
  _LOGBATCH._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _LOGMODEL._options = None
  _LOGMODEL._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _LOGINPUTS.fields_by_name['run_id']._options = None
  _LOGINPUTS.fields_by_name['run_id']._serialized_options = b'\370\206\031\001'
  _LOGINPUTS._options = None
  _LOGINPUTS._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]\342?1\n/com.databricks.mlflow.api.MlflowTrackingMessage'
  _GETEXPERIMENTBYNAME.fields_by_name['experiment_name']._options = None
  _GETEXPERIMENTBYNAME.fields_by_name['experiment_name']._serialized_options = b'\370\206\031\001'
  _GETEXPERIMENTBYNAME._options = None
  _GETEXPERIMENTBYNAME._serialized_options = b'\342?(\n&com.databricks.rpc.RPC[$this.Response]'
  _MLFLOWSERVICE.methods_by_name['getExperimentByName']._options = None
  _MLFLOWSERVICE.methods_by_name['getExperimentByName']._serialized_options = b'\362\206\031H\n,\n\003GET\022\037/mlflow/experiments/get-by-name\032\004\010\002\020\000\020\001*\026Get Experiment By Name'
  _MLFLOWSERVICE.methods_by_name['createExperiment']._options = None
  _MLFLOWSERVICE.methods_by_name['createExperiment']._serialized_options = b'\362\206\031?\n(\n\004POST\022\032/mlflow/experiments/create\032\004\010\002\020\000\020\001*\021Create Experiment'
  _MLFLOWSERVICE.methods_by_name['searchExperiments']._options = None
  _MLFLOWSERVICE.methods_by_name['searchExperiments']._serialized_options = b'\362\206\031i\n(\n\004POST\022\032/mlflow/experiments/search\032\004\010\002\020\000\n\'\n\003GET\022\032/mlflow/experiments/search\032\004\010\002\020\000\020\001*\022Search Experiments'
  _MLFLOWSERVICE.methods_by_name['getExperiment']._options = None
  _MLFLOWSERVICE.methods_by_name['getExperiment']._serialized_options = b'\362\206\0318\n$\n\003GET\022\027/mlflow/experiments/get\032\004\010\002\020\000\020\001*\016Get Experiment\272\214\031\000'
  _MLFLOWSERVICE.methods_by_name['deleteExperiment']._options = None
  _MLFLOWSERVICE.methods_by_name['deleteExperiment']._serialized_options = b'\362\206\031?\n(\n\004POST\022\032/mlflow/experiments/delete\032\004\010\002\020\000\020\001*\021Delete Experiment'
  _MLFLOWSERVICE.methods_by_name['restoreExperiment']._options = None
  _MLFLOWSERVICE.methods_by_name['restoreExperiment']._serialized_options = b'\362\206\031A\n)\n\004POST\022\033/mlflow/experiments/restore\032\004\010\002\020\000\020\001*\022Restore Experiment'
  _MLFLOWSERVICE.methods_by_name['updateExperiment']._options = None
  _MLFLOWSERVICE.methods_by_name['updateExperiment']._serialized_options = b'\362\206\031?\n(\n\004POST\022\032/mlflow/experiments/update\032\004\010\002\020\000\020\001*\021Update Experiment'
  _MLFLOWSERVICE.methods_by_name['createRun']._options = None
  _MLFLOWSERVICE.methods_by_name['createRun']._serialized_options = b'\362\206\0311\n!\n\004POST\022\023/mlflow/runs/create\032\004\010\002\020\000\020\001*\nCreate Run'
  _MLFLOWSERVICE.methods_by_name['updateRun']._options = None
  _MLFLOWSERVICE.methods_by_name['updateRun']._serialized_options = b'\362\206\0311\n!\n\004POST\022\023/mlflow/runs/update\032\004\010\002\020\000\020\001*\nUpdate Run'
  _MLFLOWSERVICE.methods_by_name['deleteRun']._options = None
  _MLFLOWSERVICE.methods_by_name['deleteRun']._serialized_options = b'\362\206\0311\n!\n\004POST\022\023/mlflow/runs/delete\032\004\010\002\020\000\020\001*\nDelete Run'
  _MLFLOWSERVICE.methods_by_name['restoreRun']._options = None
  _MLFLOWSERVICE.methods_by_name['restoreRun']._serialized_options = b'\362\206\0313\n\"\n\004POST\022\024/mlflow/runs/restore\032\004\010\002\020\000\020\001*\013Restore Run'
  _MLFLOWSERVICE.methods_by_name['logMetric']._options = None
  _MLFLOWSERVICE.methods_by_name['logMetric']._serialized_options = b'\362\206\0315\n%\n\004POST\022\027/mlflow/runs/log-metric\032\004\010\002\020\000\020\001*\nLog Metric'
  _MLFLOWSERVICE.methods_by_name['logParam']._options = None
  _MLFLOWSERVICE.methods_by_name['logParam']._serialized_options = b'\362\206\0317\n(\n\004POST\022\032/mlflow/runs/log-parameter\032\004\010\002\020\000\020\001*\tLog Param'
  _MLFLOWSERVICE.methods_by_name['setExperimentTag']._options = None
  _MLFLOWSERVICE.methods_by_name['setExperimentTag']._serialized_options = b'\362\206\031L\n4\n\004POST\022&/mlflow/experiments/set-experiment-tag\032\004\010\002\020\000\020\001*\022Set Experiment Tag'
  _MLFLOWSERVICE.methods_by_name['setTag']._options = None
  _MLFLOWSERVICE.methods_by_name['setTag']._serialized_options = b'\362\206\031/\n\"\n\004POST\022\024/mlflow/runs/set-tag\032\004\010\002\020\000\020\001*\007Set Tag'
  _MLFLOWSERVICE.methods_by_name['deleteTag']._options = None
  _MLFLOWSERVICE.methods_by_name['deleteTag']._serialized_options = b'\362\206\0315\n%\n\004POST\022\027/mlflow/runs/delete-tag\032\004\010\002\020\000\020\001*\nDelete Tag'
  _MLFLOWSERVICE.methods_by_name['getRun']._options = None
  _MLFLOWSERVICE.methods_by_name['getRun']._serialized_options = b'\362\206\031*\n\035\n\003GET\022\020/mlflow/runs/get\032\004\010\002\020\000\020\001*\007Get Run\272\214\031\000'
  _MLFLOWSERVICE.methods_by_name['searchRuns']._options = None
  _MLFLOWSERVICE.methods_by_name['searchRuns']._serialized_options = b'\362\206\0312\n!\n\004POST\022\023/mlflow/runs/search\032\004\010\002\020\000\020\001*\013Search Runs'
  _MLFLOWSERVICE.methods_by_name['listArtifacts']._options = None
  _MLFLOWSERVICE.methods_by_name['listArtifacts']._serialized_options = b'\362\206\0317\n#\n\003GET\022\026/mlflow/artifacts/list\032\004\010\002\020\000\020\001*\016List Artifacts'
  _MLFLOWSERVICE.methods_by_name['getMetricHistory']._options = None
  _MLFLOWSERVICE.methods_by_name['getMetricHistory']._serialized_options = b'\362\206\031@\n(\n\003GET\022\033/mlflow/metrics/get-history\032\004\010\002\020\000\020\001*\022Get Metric History'
  _MLFLOWSERVICE.methods_by_name['getMetricHistoryBulkInterval']._options = None
  _MLFLOWSERVICE.methods_by_name['getMetricHistoryBulkInterval']._serialized_options = b'\362\206\031:\n6\n\003GET\022)/mlflow/metrics/get-history-bulk-interval\032\004\010\002\020\013\020\003'
  _MLFLOWSERVICE.methods_by_name['logBatch']._options = None
  _MLFLOWSERVICE.methods_by_name['logBatch']._serialized_options = b'\362\206\0313\n$\n\004POST\022\026/mlflow/runs/log-batch\032\004\010\002\020\000\020\001*\tLog Batch'
  _MLFLOWSERVICE.methods_by_name['logModel']._options = None
  _MLFLOWSERVICE.methods_by_name['logModel']._serialized_options = b'\362\206\0313\n$\n\004POST\022\026/mlflow/runs/log-model\032\004\010\002\020\000\020\001*\tLog Model'
  _MLFLOWSERVICE.methods_by_name['logInputs']._options = None
  _MLFLOWSERVICE.methods_by_name['logInputs']._serialized_options = b'\362\206\0315\n%\n\004POST\022\027/mlflow/runs/log-inputs\032\004\010\002\020\000\020\001*\nLog Inputs'
  _VIEWTYPE._serialized_start=5407
  _VIEWTYPE._serialized_end=5461
  _SOURCETYPE._serialized_start=5463
  _SOURCETYPE._serialized_end=5536
  _RUNSTATUS._serialized_start=5538
  _RUNSTATUS._serialized_end=5615
  _METRIC._serialized_start=66
  _METRIC._serialized_end=138
  _PARAM._serialized_start=140
  _PARAM._serialized_end=175
  _RUN._serialized_start=177
  _RUN._serialized_end=279
  _RUNDATA._serialized_start=281
  _RUNDATA._serialized_end=384
  _RUNINPUTS._serialized_start=386
  _RUNINPUTS._serialized_end=443
  _RUNTAG._serialized_start=445
  _RUNTAG._serialized_end=481
  _EXPERIMENTTAG._serialized_start=483
  _EXPERIMENTTAG._serialized_end=526
  _RUNINFO._serialized_start=529
  _RUNINFO._serialized_end=750
  _EXPERIMENT._serialized_start=753
  _EXPERIMENT._serialized_end=940
  _DATASETINPUT._serialized_start=942
  _DATASETINPUT._serialized_end=1028
  _INPUTTAG._serialized_start=1030
  _INPUTTAG._serialized_end=1080
  _DATASET._serialized_start=1083
  _DATASET._serialized_end=1216
  _CREATEEXPERIMENT._serialized_start=1219
  _CREATEEXPERIMENT._serialized_end=1401
  _CREATEEXPERIMENT_RESPONSE._serialized_start=1323
  _CREATEEXPERIMENT_RESPONSE._serialized_end=1356
  _SEARCHEXPERIMENTS._serialized_start=1404
  _SEARCHEXPERIMENTS._serialized_end=1658
  _SEARCHEXPERIMENTS_RESPONSE._serialized_start=1537
  _SEARCHEXPERIMENTS_RESPONSE._serialized_end=1613
  _GETEXPERIMENT._serialized_start=1661
  _GETEXPERIMENT._serialized_end=1802
  _GETEXPERIMENT_RESPONSE._serialized_start=1707
  _GETEXPERIMENT_RESPONSE._serialized_end=1757
  _DELETEEXPERIMENT._serialized_start=1804
  _DELETEEXPERIMENT._serialized_end=1908
  _DELETEEXPERIMENT_RESPONSE._serialized_start=1323
  _DELETEEXPERIMENT_RESPONSE._serialized_end=1333
  _RESTOREEXPERIMENT._serialized_start=1910
  _RESTOREEXPERIMENT._serialized_end=2015
  _RESTOREEXPERIMENT_RESPONSE._serialized_start=1323
  _RESTOREEXPERIMENT_RESPONSE._serialized_end=1333
  _UPDATEEXPERIMENT._serialized_start=2017
  _UPDATEEXPERIMENT._serialized_end=2139
  _UPDATEEXPERIMENT_RESPONSE._serialized_start=1323
  _UPDATEEXPERIMENT_RESPONSE._serialized_end=1333
  _CREATERUN._serialized_start=2142
  _CREATERUN._serialized_end=2344
  _CREATERUN_RESPONSE._serialized_start=2263
  _CREATERUN_RESPONSE._serialized_end=2299
  _UPDATERUN._serialized_start=2347
  _UPDATERUN._serialized_end=2555
  _UPDATERUN_RESPONSE._serialized_start=2465
  _UPDATERUN_RESPONSE._serialized_end=2510
  _DELETERUN._serialized_start=2557
  _DELETERUN._serialized_end=2647
  _DELETERUN_RESPONSE._serialized_start=1323
  _DELETERUN_RESPONSE._serialized_end=1333
  _RESTORERUN._serialized_start=2649
  _RESTORERUN._serialized_end=2740
  _RESTORERUN_RESPONSE._serialized_start=1323
  _RESTORERUN_RESPONSE._serialized_end=1333
  _LOGMETRIC._serialized_start=2743
  _LOGMETRIC._serialized_end=2927
  _LOGMETRIC_RESPONSE._serialized_start=1323
  _LOGMETRIC_RESPONSE._serialized_end=1333
  _LOGPARAM._serialized_start=2930
  _LOGPARAM._serialized_end=3071
  _LOGPARAM_RESPONSE._serialized_start=1323
  _LOGPARAM_RESPONSE._serialized_end=1333
  _SETEXPERIMENTTAG._serialized_start=3074
  _SETEXPERIMENTTAG._serialized_end=3218
  _SETEXPERIMENTTAG_RESPONSE._serialized_start=1323
  _SETEXPERIMENTTAG_RESPONSE._serialized_end=1333
  _SETTAG._serialized_start=3221
  _SETTAG._serialized_end=3360
  _SETTAG_RESPONSE._serialized_start=1323
  _SETTAG_RESPONSE._serialized_end=1333
  _DELETETAG._serialized_start=3362
  _DELETETAG._serialized_end=3471
  _DELETETAG_RESPONSE._serialized_start=1323
  _DELETETAG_RESPONSE._serialized_end=1333
  _GETRUN._serialized_start=3473
  _GETRUN._serialized_end=3598
  _GETRUN_RESPONSE._serialized_start=2263
  _GETRUN_RESPONSE._serialized_end=2299
  _SEARCHRUNS._serialized_start=3601
  _SEARCHRUNS._serialized_end=3881
  _SEARCHRUNS_RESPONSE._serialized_start=3774
  _SEARCHRUNS_RESPONSE._serialized_end=3836
  _LISTARTIFACTS._serialized_start=3884
  _LISTARTIFACTS._serialized_end=4100
  _LISTARTIFACTS_RESPONSE._serialized_start=3969
  _LISTARTIFACTS_RESPONSE._serialized_end=4055
  _FILEINFO._serialized_start=4102
  _FILEINFO._serialized_end=4161
  _GETMETRICHISTORY._serialized_start=4164
  _GETMETRICHISTORY._serialized_end=4398
  _GETMETRICHISTORY_RESPONSE._serialized_start=4285
  _GETMETRICHISTORY_RESPONSE._serialized_end=4353
  _METRICWITHRUNID._serialized_start=4400
  _METRICWITHRUNID._serialized_end=4497
  _GETMETRICHISTORYBULKINTERVAL._serialized_start=4500
  _GETMETRICHISTORYBULKINTERVAL._serialized_end=4783
  _GETMETRICHISTORYBULKINTERVAL_RESPONSE._serialized_start=4634
  _GETMETRICHISTORYBULKINTERVAL_RESPONSE._serialized_end=4686
  _LOGBATCH._serialized_start=4786
  _LOGBATCH._serialized_end=4963
  _LOGBATCH_RESPONSE._serialized_start=1323
  _LOGBATCH_RESPONSE._serialized_end=1333
  _LOGMODEL._serialized_start=4965
  _LOGMODEL._serialized_end=5068
  _LOGMODEL_RESPONSE._serialized_start=1323
  _LOGMODEL_RESPONSE._serialized_end=1333
  _LOGINPUTS._serialized_start=5071
  _LOGINPUTS._serialized_end=5253
  _LOGINPUTS_RESPONSE._serialized_start=1323
  _LOGINPUTS_RESPONSE._serialized_end=1333
  _GETEXPERIMENTBYNAME._serialized_start=5256
  _GETEXPERIMENTBYNAME._serialized_end=5405
  _GETEXPERIMENTBYNAME_RESPONSE._serialized_start=1707
  _GETEXPERIMENTBYNAME_RESPONSE._serialized_end=1757
  _MLFLOWSERVICE._serialized_start=5618
  _MLFLOWSERVICE._serialized_end=8872
MlflowService = service_reflection.GeneratedServiceType('MlflowService', (_service.Service,), dict(
  DESCRIPTOR = _MLFLOWSERVICE,
  __module__ = 'service_pb2'
  ))

MlflowService_Stub = service_reflection.GeneratedServiceStubType('MlflowService_Stub', (MlflowService,), dict(
  DESCRIPTOR = _MLFLOWSERVICE,
  __module__ = 'service_pb2'
  ))


# @@protoc_insertion_point(module_scope)
