# This file was auto-generated by update_ml_package_versions.py.
# Please do not edit it manually.

_ML_PACKAGE_VERSIONS = {
    "sklearn": {
        "package_info": {
            "pip_release": "scikit-learn"
        },
        "models": {
            "minimum": "0.24.1",
            "maximum": "1.4.2"
        },
        "autologging": {
            "minimum": "0.24.1",
            "maximum": "1.4.2"
        }
    },
    "pytorch": {
        "package_info": {
            "pip_release": "torch"
        },
        "models": {
            "minimum": "1.9.0",
            "maximum": "2.2.2"
        },
        "autologging": {
            "minimum": "1.9.0",
            "maximum": "2.2.2"
        }
    },
    "pytorch-lightning": {
        "package_info": {
            "pip_release": "pytorch-lightning"
        },
        "autologging": {
            "minimum": "1.9.0",
            "maximum": "2.2.2"
        }
    },
    "keras": {
        "package_info": {
            "pip_release": "keras"
        },
        "models": {
            "minimum": "3.0.2",
            "maximum": "3.2.1"
        },
        "autologging": {
            "minimum": "3.0.2",
            "maximum": "3.2.1"
        }
    },
    "tensorflow": {
        "package_info": {
            "pip_release": "tensorflow"
        },
        "models": {
            "minimum": "2.6.5",
            "maximum": "2.16.1"
        },
        "autologging": {
            "minimum": "2.6.5",
            "maximum": "2.16.1"
        }
    },
    "xgboost": {
        "package_info": {
            "pip_release": "xgboost"
        },
        "models": {
            "minimum": "1.4.2",
            "maximum": "2.0.3"
        },
        "autologging": {
            "minimum": "1.4.2",
            "maximum": "2.0.3"
        }
    },
    "lightgbm": {
        "package_info": {
            "pip_release": "lightgbm"
        },
        "models": {
            "minimum": "3.1.1",
            "maximum": "4.3.0"
        },
        "autologging": {
            "minimum": "3.1.1",
            "maximum": "4.3.0"
        }
    },
    "catboost": {
        "package_info": {
            "pip_release": "catboost"
        },
        "models": {
            "minimum": "1.0.0",
            "maximum": "1.2.3"
        }
    },
    "gluon": {
        "package_info": {
            "pip_release": "mxnet"
        },
        "models": {
            "minimum": "1.5.1",
            "maximum": "1.9.1"
        },
        "autologging": {
            "minimum": "1.5.1",
            "maximum": "1.9.1"
        }
    },
    "fastai": {
        "package_info": {
            "pip_release": "fastai"
        },
        "models": {
            "minimum": "2.4.1",
            "maximum": "2.7.14"
        },
        "autologging": {
            "minimum": "2.4.1",
            "maximum": "2.7.14"
        }
    },
    "onnx": {
        "package_info": {
            "pip_release": "onnx"
        },
        "models": {
            "minimum": "1.13.1",
            "maximum": "1.16.0"
        }
    },
    "spacy": {
        "package_info": {
            "pip_release": "spacy"
        },
        "models": {
            "minimum": "2.2.4",
            "maximum": "3.7.4"
        }
    },
    "statsmodels": {
        "package_info": {
            "pip_release": "statsmodels"
        },
        "models": {
            "minimum": "0.11.1",
            "maximum": "0.14.1"
        },
        "autologging": {
            "minimum": "0.11.1",
            "maximum": "0.14.1"
        }
    },
    "spark": {
        "package_info": {
            "pip_release": "pyspark"
        },
        "models": {
            "minimum": "3.1.2",
            "maximum": "3.5.1"
        },
        "autologging": {
            "minimum": "3.1.2",
            "maximum": "3.5.1"
        }
    },
    "mleap": {
        "package_info": {
            "pip_release": "mleap"
        },
        "models": {
            "minimum": "0.18.0",
            "maximum": "0.23.1"
        }
    },
    "prophet": {
        "package_info": {
            "pip_release": "prophet"
        },
        "models": {
            "minimum": "1.1.1",
            "maximum": "1.1.5"
        }
    },
    "pmdarima": {
        "package_info": {
            "pip_release": "pmdarima"
        },
        "models": {
            "minimum": "1.8.0",
            "maximum": "2.0.4"
        }
    },
    "diviner": {
        "package_info": {
            "pip_release": "diviner"
        },
        "models": {
            "minimum": "0.1.0",
            "maximum": "0.1.1"
        }
    },
    "h2o": {
        "package_info": {
            "pip_release": "h2o"
        },
        "models": {
            "minimum": "3.40.0.1",
            "maximum": "3.46.0.1"
        }
    },
    "shap": {
        "package_info": {
            "pip_release": "shap"
        },
        "models": {
            "minimum": "0.41.0",
            "maximum": "0.45.0"
        }
    },
    "paddle": {
        "package_info": {
            "pip_release": "paddlepaddle"
        },
        "models": {
            "minimum": "2.4.1",
            "maximum": "2.6.1"
        }
    },
    "transformers": {
        "package_info": {
            "pip_release": "transformers"
        },
        "models": {
            "minimum": "4.25.1",
            "maximum": "4.39.3"
        },
        "autologging": {
            "minimum": "4.25.1",
            "maximum": "4.39.3"
        }
    },
    "openai": {
        "package_info": {
            "pip_release": "openai"
        },
        "models": {
            "minimum": "0.27.2",
            "maximum": "1.19.0"
        }
    },
    "langchain": {
        "package_info": {
            "pip_release": "langchain"
        },
        "models": {
            "minimum": "0.0.244",
            "maximum": "0.1.16"
        },
        "autologging": {
            "minimum": "0.1.4",
            "maximum": "0.1.16"
        }
    },
    "sentence_transformers": {
        "package_info": {
            "pip_release": "sentence-transformers"
        },
        "models": {
            "minimum": "2.2.2",
            "maximum": "2.6.1"
        }
    },
    "johnsnowlabs": {
        "package_info": {
            "pip_release": "johnsnowlabs"
        },
        "models": {
            "minimum": "4.4.6",
            "maximum": "5.3.4"
        }
    },
    "promptflow": {
        "package_info": {
            "pip_release": "promptflow"
        },
        "models": {
            "minimum": "1.3.0",
            "maximum": "1.8.0"
        }
    }
}
